@@ -13,7 +13,7 @@ import (
 	"github.com/alecthomas/chroma/v2/lexers"
 	"github.com/alecthomas/chroma/v2/styles"
 	"github.com/google/go-cmp/cmp"
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
 	"github.com/walles/moar/m/textstyles"
 	"github.com/walles/moar/twin"
 	"gotest.tools/v3/assert"
@@ -94,7 +94,7 @@ func TestBrokenUtf8(t *testing.T) {
 	}
 }
 
-func startPaging(t *testing.T, reader *Reader) *twin.FakeScreen {
+func startPaging(t *testing.T, reader *ReaderImpl) *twin.FakeScreen {
 	err := reader._wait()
 	if err != nil {
 		t.Fatalf("Failed waiting for reader: %v", err)
@@ -117,7 +117,7 @@ func startPaging(t *testing.T, reader *Reader) *twin.FakeScreen {
 }
 
 // Set style to "native" and use the TTY16m formatter
-func startPagingWithTerminalFg(t *testing.T, reader *Reader, withTerminalFg bool) *twin.FakeScreen {
+func startPagingWithTerminalFg(t *testing.T, reader *ReaderImpl, withTerminalFg bool) *twin.FakeScreen {
 	err := reader._wait()
 	if err != nil {
 		t.Fatalf("Failed waiting for reader: %v", err)
@@ -336,8 +336,8 @@ func TestFindFirstHitSimple(t *testing.T) {
 
 	pager.searchPattern = toPattern("AB")
 
-	hit := pager.findFirstHit(linenumbers.LineNumber{}, nil, false)
-	assert.Assert(t, hit.internalDontTouch.lineNumber.IsZero())
+	hit := pager.findFirstHit(linemetadata.Index{}, nil, false)
+	assert.Assert(t, hit.internalDontTouch.lineIndex.IsZero())
 	assert.Equal(t, hit.internalDontTouch.deltaScreenLines, 0)
 }
 
@@ -350,8 +350,8 @@ func TestFindFirstHitAnsi(t *testing.T) {
 
 	pager.searchPattern = toPattern("AB")
 
-	hit := pager.findFirstHit(linenumbers.LineNumber{}, nil, false)
-	assert.Assert(t, hit.internalDontTouch.lineNumber.IsZero())
+	hit := pager.findFirstHit(linemetadata.Index{}, nil, false)
+	assert.Assert(t, hit.internalDontTouch.lineIndex.IsZero())
 	assert.Equal(t, hit.internalDontTouch.deltaScreenLines, 0)
 }
 
@@ -364,7 +364,7 @@ func TestFindFirstHitNoMatch(t *testing.T) {
 
 	pager.searchPattern = toPattern("this pattern should not be found")
 
-	hit := pager.findFirstHit(linenumbers.LineNumber{}, nil, false)
+	hit := pager.findFirstHit(linemetadata.Index{}, nil, false)
 	assert.Assert(t, hit == nil)
 }
 
@@ -376,7 +376,7 @@ func TestFindFirstHitNoMatchBackwards(t *testing.T) {
 	assert.NilError(t, pager.reader._wait())
 
 	pager.searchPattern = toPattern("this pattern should not be found")
-	theEnd := *linenumbers.LineNumberFromLength(reader.GetLineCount())
+	theEnd := *linemetadata.IndexFromLength(reader.GetLineCount())
 
 	hit := pager.findFirstHit(theEnd, nil, true)
 	assert.Assert(t, hit == nil)
@@ -583,7 +583,7 @@ func TestPageSamples(t *testing.T) {
 			pager.StartPaging(screen, nil, nil)
 			pager.redraw("")
 
-			firstReaderLine := myReader.GetLine(linenumbers.LineNumber{})
+			firstReaderLine := myReader.GetLine(linemetadata.Index{})
 			if firstReaderLine == nil {
 				return
 			}
@@ -593,9 +593,9 @@ func TestPageSamples(t *testing.T) {
 			firstPagerLine = strings.TrimSuffix(firstPagerLine, ">")
 
 			assert.Assert(t,
-				strings.HasPrefix(firstReaderLine.Plain(nil), firstPagerLine),
+				strings.HasPrefix(firstReaderLine.Plain(), firstPagerLine),
 				"\nreader line = <%s>\npager line  = <%s>",
-				firstReaderLine.Plain(nil), firstPagerLine,
+				firstReaderLine.Plain(), firstPagerLine,
 			)
 		})
 	}
@@ -743,7 +743,7 @@ func benchmarkSearch(b *testing.B, highlighted bool) {
 	b.ResetTimer()
 
 	// This test will search through all the N copies we made of our file
-	hit := pager.findFirstHit(linenumbers.LineNumber{}, nil, false)
+	hit := pager.findFirstHit(linemetadata.Index{}, nil, false)
 
 	if hit != nil {
 		panic(fmt.Errorf("This test is meant to scan the whole file without finding anything"))
