@@ -3,7 +3,7 @@ package m
 import (
 	"fmt"
 
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
 )
 
 // Please create using newScrollPosition(name)
@@ -23,9 +23,9 @@ func newScrollPosition(name string) scrollPosition {
 }
 
 type scrollPositionInternal struct {
-	// Line number in the input stream, or nil if nothing has been read yet or
-	// there are no lines.
-	lineNumber *linenumbers.LineNumber
+	// Index into the array of visible input lines, or nil if nothing has been
+	// read yet or there are no lines.
+	lineIndex *linemetadata.Index
 
 	// Scroll this many screen lines before rendering. Can be negative.
 	deltaScreenLines int
@@ -43,8 +43,10 @@ type scrollPositionCanonical struct {
 	showStatusBar   bool // From pager
 	wrapLongLines   bool // From pager
 
-	lineNumber       *linenumbers.LineNumber // From scrollPositionInternal
-	deltaScreenLines int                     // From scrollPositionInternal
+	pagerLineCount int // From pager.Reader().GetLineCount()
+
+	lineIndex        *linemetadata.Index // From scrollPositionInternal
+	deltaScreenLines int                 // From scrollPositionInternal
 }
 
 func canonicalFromPager(pager *Pager) scrollPositionCanonical {
@@ -56,7 +58,9 @@ func canonicalFromPager(pager *Pager) scrollPositionCanonical {
 		showStatusBar:   pager.ShowStatusBar,
 		wrapLongLines:   pager.WrapLongLines,
 
-		lineNumber:       pager.scrollPosition.internalDontTouch.lineNumber,
+		pagerLineCount: pager.Reader().GetLineCount(),
+
+		lineIndex:        pager.scrollPosition.internalDontTouch.lineIndex,
 		deltaScreenLines: pager.scrollPosition.internalDontTouch.deltaScreenLines,
 	}
 }
@@ -66,7 +70,7 @@ func (sp scrollPosition) PreviousLine(scrollDistance int) scrollPosition {
 	return scrollPosition{
 		internalDontTouch: scrollPositionInternal{
 			name:             sp.internalDontTouch.name,
-			lineNumber:       sp.internalDontTouch.lineNumber,
+			lineIndex:        sp.internalDontTouch.lineIndex,
 			deltaScreenLines: sp.internalDontTouch.deltaScreenLines - scrollDistance,
 		},
 	}
@@ -77,7 +81,7 @@ func (sp scrollPosition) NextLine(scrollDistance int) scrollPosition {
 	return scrollPosition{
 		internalDontTouch: scrollPositionInternal{
 			name:             sp.internalDontTouch.name,
-			lineNumber:       sp.internalDontTouch.lineNumber,
+			lineIndex:        sp.internalDontTouch.lineIndex,
 			deltaScreenLines: sp.internalDontTouch.deltaScreenLines + scrollDistance,
 		},
 	}
@@ -86,11 +90,11 @@ func (sp scrollPosition) NextLine(scrollDistance int) scrollPosition {
 // Create a new position, scrolled to the given line number
 //
 //revive:disable-next-line:unexported-return
-func NewScrollPositionFromLineNumber(lineNumber linenumbers.LineNumber, name string) scrollPosition {
+func NewScrollPositionFromIndex(index linemetadata.Index, name string) scrollPosition {
 	return scrollPosition{
 		internalDontTouch: scrollPositionInternal{
 			name:             name,
-			lineNumber:       &lineNumber,
+			lineIndex:        &index,
 			deltaScreenLines: 0,
 		},
 	}
@@ -98,19 +102,19 @@ func NewScrollPositionFromLineNumber(lineNumber linenumbers.LineNumber, name str
 
 // Move towards the top until deltaScreenLines is not negative any more
 func (si *scrollPositionInternal) handleNegativeDeltaScreenLines(pager *Pager) {
-	for !si.lineNumber.IsZero() && si.deltaScreenLines < 0 {
+	for !si.lineIndex.IsZero() && si.deltaScreenLines < 0 {
 		// Render the previous line
-		previousLineNumber := si.lineNumber.NonWrappingAdd(-1)
-		previousLine := pager.reader.GetLine(previousLineNumber)
-		previousSubLines := pager.renderLine(previousLine, previousLineNumber, *si)
+		previousLineIndex := si.lineIndex.NonWrappingAdd(-1)
+		previousLine := pager.Reader().GetLine(previousLineIndex)
+		previousSubLines := pager.renderLine(previousLine, si.getMaxNumberPrefixLength(pager))
 
 		// Adjust lineNumber and deltaScreenLines to move up into the previous
 		// screen line
-		si.lineNumber = &previousLineNumber
+		si.lineIndex = &previousLineIndex
 		si.deltaScreenLines += len(previousSubLines)
 	}
 
-	if si.lineNumber.IsZero() && si.deltaScreenLines <= 0 {
+	if si.lineIndex.IsZero() && si.deltaScreenLines <= 0 {
 		// Can't go any higher
 		si.deltaScreenLines = 0
 		return
@@ -123,30 +127,24 @@ func (si *scrollPositionInternal) handleNegativeDeltaScreenLines(pager *Pager) {
 // This method will not do any screen-height based clipping, so it could be that
 // the position is too far down to display after this returns.
 func (si *scrollPositionInternal) handlePositiveDeltaScreenLines(pager *Pager) {
-	for {
-		line := pager.reader.GetLine(*si.lineNumber)
-		if line == nil {
-			// Out of bounds downwards, get the last line...
-			si.lineNumber = linenumbers.LineNumberFromLength(pager.reader.GetLineCount())
-			line = pager.reader.GetLine(*si.lineNumber)
-			if line == nil {
-				panic(fmt.Errorf("Last line is nil"))
-			}
-			subLines := pager.renderLine(line, *si.lineNumber, *si)
+	maxPrefixLength := 0
+	allPossibleLines := pager.Reader().GetLines(*si.lineIndex, pager.visibleHeight())
+	if len(allPossibleLines.lines) > 0 {
+		lastPossibleLine := allPossibleLines.lines[len(allPossibleLines.lines)-1]
+		maxPrefixLength = pager.getLineNumberPrefixLength(lastPossibleLine.number)
+	}
 
-			// ... and go to the bottom of that.
-			si.deltaScreenLines = len(subLines) - 1
-			return
-		}
+	for _, line := range allPossibleLines.lines {
+		si.lineIndex = &line.index
 
-		subLines := pager.renderLine(line, *si.lineNumber, *si)
+		subLines := pager.renderLine(line, maxPrefixLength)
 		if si.deltaScreenLines < len(subLines) {
 			// Sublines are within bounds!
 			return
 		}
 
-		nextLineNumber := si.lineNumber.NonWrappingAdd(1)
-		si.lineNumber = &nextLineNumber
+		nextLineIndex := si.lineIndex.NonWrappingAdd(1)
+		si.lineIndex = &nextLineIndex
 		si.deltaScreenLines -= len(subLines)
 	}
 }
@@ -154,34 +152,50 @@ func (si *scrollPositionInternal) handlePositiveDeltaScreenLines(pager *Pager) {
 // This method assumes si contains a canonical position
 func (si *scrollPositionInternal) emptyBottomLinesCount(pager *Pager) int {
 	unclaimedViewportLines := pager.visibleHeight()
+	if pager.Reader().GetLineCount() == 0 {
+		// No lines available, so all viewport lines are unclaimed
+		return unclaimedViewportLines
+	}
 
 	// Start counting where the current input line begins
 	unclaimedViewportLines += si.deltaScreenLines
 
-	lineNumber := *si.lineNumber
+	lineIndex := *si.lineIndex
+
+	var lastLine NumberedLine
+	lastLineIndex := linemetadata.IndexFromLength(pager.Reader().GetLineCount())
+	if lastLineIndex != nil {
+		maybeLastLine := pager.Reader().GetLine(*lastLineIndex)
+		// This check is needed for the unlikely case that we just reformatted
+		// the input stream and it just lost some lines.
+		if maybeLastLine != nil {
+			lastLine = *maybeLastLine
+		}
+	}
+	lastLineNumberWidth := pager.getLineNumberPrefixLength(lastLine.number)
 
 	for {
-		line := pager.reader.GetLine(lineNumber)
+		line := pager.Reader().GetLine(lineIndex)
 		if line == nil {
 			// No more lines!
 			break
 		}
 
-		subLines := pager.renderLine(line, lineNumber, *si)
+		subLines := pager.renderLine(line, lastLineNumberWidth)
 		unclaimedViewportLines -= len(subLines)
 		if unclaimedViewportLines <= 0 {
 			return 0
 		}
 
 		// Move to the next line
-		lineNumber = lineNumber.NonWrappingAdd(1)
+		lineIndex = lineIndex.NonWrappingAdd(1)
 	}
 
 	return unclaimedViewportLines
 }
 
 func (si *scrollPositionInternal) isCanonical(pager *Pager) bool {
-	if si.canonical.lineNumber == nil {
+	if si.canonical.lineIndex == nil {
 		// Awaiting initial lines from the reader
 		return false
 	}
@@ -199,13 +213,13 @@ func (sp scrollPosition) isVisible(pager *Pager) bool {
 		panic(fmt.Errorf("Negative incoming deltaScreenLines: %#v", sp.internalDontTouch))
 	}
 
-	if sp.internalDontTouch.lineNumber.IsBefore(*pager.lineNumber()) {
+	if sp.internalDontTouch.lineIndex.IsBefore(*pager.lineIndex()) {
 		// Line number too low, not visible
 		return false
 	}
 
 	lastVisiblePosition := pager.getLastVisiblePosition()
-	if sp.internalDontTouch.lineNumber.IsAfter(*lastVisiblePosition.lineNumber(pager)) {
+	if sp.internalDontTouch.lineIndex.IsAfter(*lastVisiblePosition.lineIndex(pager)) {
 		// Line number too high, not visible
 		return false
 	}
@@ -240,15 +254,15 @@ func (si *scrollPositionInternal) canonicalize(pager *Pager) {
 		si.canonicalizing = false
 	}()
 
-	if pager.reader.GetLineCount() == 0 {
-		si.lineNumber = nil
+	if pager.Reader().GetLineCount() == 0 {
+		si.lineIndex = nil
 		si.deltaScreenLines = 0
 		return
 	}
 
-	if si.lineNumber == nil {
+	if si.lineIndex == nil {
 		// We have lines, but no line number, start at the top
-		si.lineNumber = &linenumbers.LineNumber{}
+		si.lineIndex = &linemetadata.Index{}
 	}
 
 	si.handleNegativeDeltaScreenLines(pager)
@@ -263,25 +277,25 @@ func (si *scrollPositionInternal) canonicalize(pager *Pager) {
 	}
 }
 
-func scrollPositionFromLineNumber(name string, lineNumber linenumbers.LineNumber) *scrollPosition {
+func scrollPositionFromIndex(name string, index linemetadata.Index) *scrollPosition {
 	return &scrollPosition{
 		internalDontTouch: scrollPositionInternal{
-			name:       name,
-			lineNumber: &lineNumber,
+			name:      name,
+			lineIndex: &index,
 		},
 	}
 }
 
-// Line number in the input stream, or nil if nothing has been read
-func (p *Pager) lineNumber() *linenumbers.LineNumber {
+// Line index in the input stream, or nil if nothing has been read
+func (p *Pager) lineIndex() *linemetadata.Index {
 	p.scrollPosition.internalDontTouch.canonicalize(p)
-	return p.scrollPosition.internalDontTouch.lineNumber
+	return p.scrollPosition.internalDontTouch.lineIndex
 }
 
-// Line number in the input stream, or nil if nothing has been read
-func (sp *scrollPosition) lineNumber(pager *Pager) *linenumbers.LineNumber {
+// Line index in the input stream, or nil if nothing has been read
+func (sp *scrollPosition) lineIndex(pager *Pager) *linemetadata.Index {
 	sp.internalDontTouch.canonicalize(pager)
-	return sp.internalDontTouch.lineNumber
+	return sp.internalDontTouch.lineIndex
 }
 
 // Scroll this many screen lines before rendering
@@ -301,51 +315,51 @@ func (sp *scrollPosition) deltaScreenLines(pager *Pager) int {
 }
 
 func (p *Pager) scrollToEnd() {
-	inputLineCount := p.reader.GetLineCount()
+	inputLineCount := p.Reader().GetLineCount()
 	if inputLineCount == 0 {
 		return
 	}
-	lastInputLineNumber := *linenumbers.LineNumberFromLength(inputLineCount)
+	lastInputIndex := *linemetadata.IndexFromLength(inputLineCount)
 
-	lastInputLine := p.reader.GetLine(lastInputLineNumber)
+	lastInputLine := p.Reader().GetLine(lastInputIndex)
 
-	p.scrollPosition.internalDontTouch.lineNumber = &lastInputLineNumber
+	p.scrollPosition.internalDontTouch.lineIndex = &lastInputIndex
 
 	// Scroll down enough. We know for sure the last line won't wrap into more
 	// lines than the number of characters it contains.
-	p.scrollPosition.internalDontTouch.deltaScreenLines = len(lastInputLine.raw)
+	p.scrollPosition.internalDontTouch.deltaScreenLines = len(lastInputLine.line.raw)
 
-	if p.TargetLineNumber == nil {
+	if p.TargetLine == nil {
 		// Start following the end of the file
 		//
 		// Otherwise, if we're already aiming for some place, don't overwrite
 		// that.
-		maxLineNumber := linenumbers.LineNumberMax()
-		p.TargetLineNumber = &maxLineNumber
+		maxLineIndex := linemetadata.IndexMax()
+		p.TargetLine = &maxLineIndex
 	}
 }
 
 // Can be either because Pager.scrollToEnd() was just called or because the user
 // has pressed the down arrow enough times.
 func (p *Pager) isScrolledToEnd() bool {
-	inputLineCount := p.reader.GetLineCount()
+	inputLineCount := p.Reader().GetLineCount()
 	if inputLineCount == 0 {
 		// No lines available, which means we can't scroll any further down
 		return true
 	}
-	lastInputLineNumber := *linenumbers.LineNumberFromLength(inputLineCount)
+	lastInputLineIndex := *linemetadata.IndexFromLength(inputLineCount)
 
 	visibleLines, _ := p.renderLines()
 	lastVisibleLine := visibleLines[len(visibleLines)-1]
-	if lastVisibleLine.inputLine != lastInputLineNumber {
+	if lastVisibleLine.inputLineIndex != lastInputLineIndex {
 		// Last input line is not on the screen
 		return false
 	}
 
 	// Last line is on screen, now we need to figure out whether we can see all
 	// of it
-	lastInputLine := p.reader.GetLine(lastInputLineNumber)
-	lastInputLineRendered := p.renderLine(lastInputLine, lastInputLineNumber, p.scrollPosition.internalDontTouch)
+	lastInputLine := p.Reader().GetLine(lastInputLineIndex)
+	lastInputLineRendered := p.renderLine(lastInputLine, p.getLineNumberPrefixLength(lastInputLine.number))
 	lastRenderedSubLine := lastInputLineRendered[len(lastInputLineRendered)-1]
 
 	// If the last visible subline is the same as the last possible subline then
@@ -364,21 +378,14 @@ func (p *Pager) getLastVisiblePosition() *scrollPosition {
 	return &scrollPosition{
 		internalDontTouch: scrollPositionInternal{
 			name:             "Last Visible Position",
-			lineNumber:       &lastRenderedLine.inputLine,
+			lineIndex:        &lastRenderedLine.inputLineIndex,
 			deltaScreenLines: lastRenderedLine.wrapIndex,
 		},
 	}
 }
 
-func numberPrefixLength(pager *Pager, scrollPosition scrollPositionInternal) int {
-	// This method used to live in screenLines.go, but I moved it here because
-	// it touches scroll position internals.
-
-	if !pager.ShowLineNumbers {
-		return 0
-	}
-
-	maxPossibleLineNumber := *linenumbers.LineNumberFromLength(pager.reader.GetLineCount())
+func (si *scrollPositionInternal) getMaxNumberPrefixLength(pager *Pager) int {
+	maxPossibleIndex := *linemetadata.IndexFromLength(pager.Reader().GetLineCount())
 
 	// This is an approximation assuming we don't do any wrapping. Finding the
 	// real answer while wrapping requires rendering, which requires the real
@@ -386,26 +393,26 @@ func numberPrefixLength(pager *Pager, scrollPosition scrollPositionInternal) int
 	// recursion.
 	//
 	// Let's improve on demand.
-	var lineNumber linenumbers.LineNumber
+	var index linemetadata.Index
 	// Ref: https://github.com/walles/moar/issues/198
-	if scrollPosition.lineNumber != nil {
-		lineNumber = *scrollPosition.lineNumber
+	if si.lineIndex != nil {
+		index = *si.lineIndex
 	}
-	maxVisibleLineNumber := lineNumber.NonWrappingAdd(
-		scrollPosition.deltaScreenLines +
+	maxVisibleIndex := index.NonWrappingAdd(
+		si.deltaScreenLines +
 			pager.visibleHeight() - 1)
-	if maxVisibleLineNumber.IsAfter(maxPossibleLineNumber) {
-		maxVisibleLineNumber = maxPossibleLineNumber
+	if maxVisibleIndex.IsAfter(maxPossibleIndex) {
+		maxVisibleIndex = maxPossibleIndex
+	}
+
+	var number linemetadata.Number
+	lastVisibleLine := pager.Reader().GetLine(maxVisibleIndex)
+
+	// nil can happen when the input stream is empty
+	if lastVisibleLine != nil {
+		number = lastVisibleLine.number
 	}
 
 	// Count the length of the last line number
-	numberPrefixLength := len(maxVisibleLineNumber.Format()) + 1
-	if numberPrefixLength < 4 {
-		// 4 = space for 3 digits followed by one whitespace
-		//
-		// https://github.com/walles/moar/issues/38
-		numberPrefixLength = 4
-	}
-
-	return numberPrefixLength
+	return pager.getLineNumberPrefixLength(number)
 }
