@@ -9,6 +9,7 @@ import (
 	"runtime/debug"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 
 	"github.com/alecthomas/chroma/v2"
@@ -19,7 +20,7 @@ import (
 	"golang.org/x/term"
 
 	"github.com/walles/moar/m"
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
 	"github.com/walles/moar/m/textstyles"
 	"github.com/walles/moar/twin"
 )
@@ -229,7 +230,7 @@ func tryOpen(filename string) error {
 // line number, and returns the remaining args.
 //
 // Returns nil on no target line number specified.
-func getTargetLineNumber(args []string) (*linenumbers.LineNumber, []string) {
+func getTargetLine(args []string) (*linemetadata.Index, []string) {
 	for i, arg := range args {
 		if !strings.HasPrefix(arg, "+") {
 			continue
@@ -252,7 +253,7 @@ func getTargetLineNumber(args []string) (*linenumbers.LineNumber, []string) {
 		remainingArgs = append(remainingArgs, args[:i]...)
 		remainingArgs = append(remainingArgs, args[i+1:]...)
 
-		returnMe := linenumbers.LineNumberFromOneBased(int(lineNumber))
+		returnMe := linemetadata.IndexFromOneBased(int(lineNumber))
 		return &returnMe, remainingArgs
 	}
 
@@ -388,7 +389,7 @@ func pagerFromArgs(
 		flags = append(strings.Fields(moarEnv), flags...)
 	}
 
-	targetLineNumber, remainingArgs := getTargetLineNumber(flags)
+	targetLine, remainingArgs := getTargetLine(flags)
 
 	err = flagSet.Parse(remainingArgs)
 	if err != nil {
@@ -480,15 +481,16 @@ func pagerFromArgs(
 	}
 
 	formatter := formatters.TTY256
-	if *terminalColorsCount == twin.ColorCount8 {
+	switch *terminalColorsCount {
+	case twin.ColorCount8:
 		formatter = formatters.TTY8
-	} else if *terminalColorsCount == twin.ColorCount16 {
+	case twin.ColorCount16:
 		formatter = formatters.TTY16
-	} else if *terminalColorsCount == twin.ColorCount24bit {
+	case twin.ColorCount24bit:
 		formatter = formatters.TTY16m
 	}
 
-	var reader *m.Reader
+	var reader *m.ReaderImpl
 	shouldFormat := *reFormat
 	if stdinIsRedirected {
 		// Display input pipe contents
@@ -573,17 +575,36 @@ func pagerFromArgs(
 	pager.ScrollRightHint = *scrollRightHint
 	pager.SideScrollAmount = int(*shift)
 
-	pager.TargetLineNumber = targetLineNumber
-	if *follow && pager.TargetLineNumber == nil {
-		reallyHigh := linenumbers.LineNumberMax()
-		pager.TargetLineNumber = &reallyHigh
+	pager.TargetLine = targetLine
+	if *follow && pager.TargetLine == nil {
+		reallyHigh := linemetadata.IndexMax()
+		pager.TargetLine = &reallyHigh
 	}
 
 	return pager, screen, style, &formatter, logsRequested, nil
 }
 
+type logWriter struct {
+	lock   sync.Mutex
+	buffer strings.Builder
+}
+
+func (lw *logWriter) Write(p []byte) (n int, err error) {
+	lw.lock.Lock()
+	defer lw.lock.Unlock()
+
+	return lw.buffer.Write(p)
+}
+
+func (lw *logWriter) String() string {
+	lw.lock.Lock()
+	defer lw.lock.Unlock()
+
+	return lw.buffer.String()
+}
+
 func main() {
-	var loglines strings.Builder
+	var loglines logWriter
 	logsRequested := false
 	log.SetOutput(&loglines)
 	russiaNotSupported()
